BASKET_NETWORK_FAILURE = 1
BASKET_INVALID_EMAIL = 2
BASKET_UNKNOWN_EMAIL = 3
BASKET_UNKNOWN_TOKEN = 4
BASKET_USAGE_ERROR = 5
BASKET_EMAIL_PROVIDER_AUTH_FAILURE = 6
BASKET_AUTH_ERROR = 7
BASKET_SSL_REQUIRED = 8
BASKET_INVALID_NEWSLETTER = 9
BASKET_INVALID_LANGUAGE = 10
BASKET_EMAIL_NOT_CHANGED = 11
BASKET_CHANGE_REQUEST_NOT_FOUND = 12
BASKET_MOCK_FAILURE = 13

# If you get this, report it as a bug so we can add a more specific
# error code.
BASKET_UNKNOWN_ERROR = 99
